#ifndef __LINBOX_DOMAIN_INTEGER_SMITH_NORMAL_FORM_C__
#define __LINBOX_DOMAIN_INTEGER_SMITH_NORMAL_FORM_C__
// ======================================================================= //
// Copyright (C) LinBox project 1999
// Givaro / Athapascan-1
// Smith form
// Time-stamp: <08 Dec 11 12:13:36 Jean-Guillaume.Dumas@imag.fr> 
// ======================================================================= //

#include <stdlib.h>
#include <stdio.h>

#include <givgfq.h>
#include <givintfactor.h>
#define MAXGFQSIZE 65536
#define NORMGFQSIZE 65536
#define MINGFQSIZE 32678
typedef GFqDom<long>               WGFqDomain;
typedef GFqDomain::element      WResidu;


#include "LinBox/lin_rand.h"     // Symetric Wiedemann
#include "LinBox/lin_massey.C"     // Symetric Wiedemann
#include "LinBox/lin_spv_bb.h"    // givaro sparse matrices
#include "LinBox/lin_erase_bb.h"    // givaro sparse matrices
#include "LinBox/lin_wrap_c++.h"    // givaro sparse matrices
typedef SparseBlackBoxEraseDom< WGFqDomain > WSPBBE;
typedef SparseBlackBoxDom< OperatorWrapper< WGFqDomain::Residu_t> > WSPBBlong;




void wiedemann_rank(ulong & rank, char * File_Name, ulong MOD, double &wrong) {
    startActivity(GlobalCommentator, "Wiedemann", "Computing rank via Wiedemann algorithm", LVL_NORMAL, INTERNAL_DESCRIPTION); 

    unsigned long ni(0), nj(0);
    WSPBBlong::Domain_t LongDom;
    WSPBBlong TheMat( LongDom );
    TheMat.read(ni,nj, File_Name);
    unsigned long sq = ni*nj;

    unsigned long expo = 0, qMOD = MOD, pMOD = 1;
    
    
    while ((qMOD <= MAXGFQSIZE) && ( (pMOD <= 50*sq) || (qMOD <= NORMGFQSIZE) ) ) {
        ++expo;
        pMOD = qMOD;
        qMOD *= MOD;
// cerr << "e: " << expo << ", p: " << pMOD << ", q: " << qMOD << endl;
    }    
    WGFqDomain F(MOD,expo);

    char * UT = new char[200]; 
    sprintf(UT,"Computing rank mod %ld in GF(%ld).",MOD,pMOD); 
    activityReport(GlobalCommentator, UT, LVL_UNIMP, INTERNAL_DESCRIPTION ); 


    WSPBBE MF( F );
   

    unsigned long addrank = MF.read_erase(File_Name);
    if (addrank > 0) {
        sprintf(UT,"Reordering, ... matrix reduced, preliminary rank found : %ld.", addrank);
        activityReport(GlobalCommentator, UT, LVL_UNIMP, INTERNAL_DESCRIPTION ); 
        activityReport(GlobalCommentator, "Now computing the rank of the remaining matrix.", LVL_UNIMP, INTERNAL_DESCRIPTION );
    }
    int level;
    if ( (ni > 0) && (nj > 0)) {
        Random generator;
        typedef BB_Symmetrize_Container< WSPBBE > SzCBB;
        MF.precondition(generator);
        SzCBB TF( &MF, generator); 
        MasseyDom< SzCBB >  WD(GlobalCommentator, &TF );
        WD.pseudo_rank(rank);
        
     
//         if ( (sq = MF.n_row()*MF.n_col() ) > qMOD )
//             cerr << "sq: " << sq << ", whereas field size is : " << F.size() << endl;
        rank += addrank;
    
        if ( (6*sq) > pMOD ) {
            level = LVL_ALWAYS;
            sprintf(UT,"Rank %ld, computed via a heuristic, PROBABILITY OF SUCCESS UNKNOWN.", rank); 
            wrong = 1;
        } else {
            level = LVL_IMP;
            sprintf(UT,"Rank %ld, correct with probability at least 1 - %e", rank, wrong = double(5.5*sq+1)/double(pMOD));
        }
    } else {
        level = LVL_NORMAL;
        sprintf(UT,"Rank %ld, computed deterministically.", rank = addrank); 
    }    
    stopActivity(GlobalCommentator, UT, level, INTERNAL_DESCRIPTION ); 
    delete [] UT;
}

template<class InvClass>
void wiedemann_rank_N_torsion(InvClass& Inv, char * File_Name, Integer IN) {
    startActivity(GlobalCommentator, "N-Torsion", "Rank N-Torsion Computation via wiedemann algorithm", LVL_NORMAL, INTERNAL_DESCRIPTION); 

    Integer valence;
    ulong nbfactor;
    std::list<Integer> Lm;
    std::list<ulong> Le;

    IntFactorDom<> IFD;
    nbfactor = IFD.set(Lm,Le,IN);

    typename std::list<Integer>::iterator im=Lm.begin();
    typename std::list<ulong>::iterator ie=Le.begin();

    unsigned long ni, nj;
    FILE* FileDes = fopen(File_Name, "r");
    if (FileDes != NULL) {
        fscanf(FileDes,"%ld %ld M\n",&ni, &nj) ;
    }
    fclose(FileDes);
    char * UT = new char[170]; 

    unsigned long * r = new unsigned long[nbfactor];
    bool * t = new bool[nbfactor];
    unsigned long ir;
    unsigned long k, minr= GIVMIN(ni,nj);
    Integer pr=1;
    for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
        if (*ie > 0) {
            pr *= *im;
            if (*im < Integer(MAXGFQSIZE)) {
                t[k] = true;
                double rank_wrong;
                wiedemann_rank(r[k], File_Name, (long)(*im), rank_wrong);
                minr = GIVMIN(minr,r[k]);
                sprintf(UT,"rank mod %ld : %ld", (long)(*im), r[k]);
            } else {
                t[k] = false;
                std::string s; Integer2string(s, *im);
                sprintf(UT,"WARNING, %s is TOO BIG. Supposed full rank, WARNING.", s.c_str());
            }
            
            activityReport(GlobalCommentator, UT, LVL_NORMAL, INTERNAL_WARNING);
        }
    
    Integer p; IFD.prevprime(p, MAXGFQSIZE - (ulong)lrand48()%(ulong)MAXGFQSIZE/2);
    

    while ( iszero(pr % p)) IFD.prevprimein(p);
    double rank_wrong;
//     cerr << (long)p << endl;
    
    wiedemann_rank(ir, File_Name,(long)p,rank_wrong);
    sprintf(UT,"Integer rank (mod %ld) : %ld", (long)(p), ir);
    activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
    if (ir > minr) {
        unsigned long ssize = ir-minr;
        ulong * Smith = new ulong[ssize];
        unsigned long ll=ssize;
        for (;ll--;)
            Smith[ll] = 1;
    

        unsigned long l=0, count_ie;
        for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
            if (t[k]) {
                if (*ie > 0) {
                    count_ie = 0;
                    l = r[k];
                    ulong p = (long)(*im);
                    if (l < ir) {
                        Smith[ssize-ir+l] *= p;
                        sprintf(UT,"Rank mod powers is NOT YET IMPLEMENTED, so rank mod %ld should be checked.", (Smith[ssize-ir+l]*p));
                        activityReport(GlobalCommentator, UT, LVL_NORMAL, INTERNAL_DESCRIPTION);
                    }
                }
            }

        Inv.PushInvariant(1,minr);
        unsigned long lj=0;
        for(ll=0;ll<ssize;) {
            while ((++lj<ssize) && (Smith[lj] == Smith[ll])) {};
            Inv.PushInvariant(Smith[ll],lj-ll);
            ll = lj;
        }
        delete [] Smith;
    
    } else {
        Inv.PushInvariant(1,ir);
    }
    
    delete [] UT;
    delete [] r;
    delete [] t;
    stopActivity(GlobalCommentator, "Torsions computed via a heuristic, probability of success unknown.", LVL_ALWAYS, INTERNAL_DESCRIPTION ); 
};



    
void wiedemann_Integer_rank(ulong& rank, Integer& p1, char * File_Name, Integer IN, double& wrong, double error) {
    IntPrimeDom IPD;
    IPD.prevprime(p1, 1+MAXGFQSIZE - (ulong)lrand48()%(ulong)MAXGFQSIZE/20);
    Integer Prod = IN;
    
    
    while ( IPD.iszero(Prod % p1)) IPD.prevprimein(p1);

    double rank_wrong, int_wrong;
    wiedemann_rank(rank, File_Name, (long)p1, rank_wrong);

    if (rank_wrong < 1) {
            // Taking error into account ...
        int nbr = 0;
        if (error == 0) error = MINERRORLEVEL;
        while (error <= (int_wrong = rank_wrong+wrong)){ 
            double r_w = 0;
            Prod *= p1;
            while ( IPD.iszero(Prod % p1)) IPD.prevprimein(p1);
            unsigned long rr;
            wiedemann_rank(rr, File_Name, (long)p1, r_w);
            if (rr > rank) {        // Previous ranks were bad
                nbr = 0;
                rank_wrong = r_w;
                rank = rr;
            } else if (rr == rank) { // another rank
                rank_wrong = (r_w)*(rank_wrong);
            } else               // bad rank (happens with proba wrong)
                --nbr;
        }
    } else {
            // else is impossible : heuristic did not gave probability.
            // We therefore just do it twice.
        int nbr = 0;
        while ( ++nbr < 2 ){ 
            double r_w = 0;
            Prod *= p1;
            while ( IPD.iszero(Prod % p1)) IPD.prevprimein(p1);
            unsigned long rr;
            wiedemann_rank(rr, File_Name, (long)p1, r_w);
            if (rr > rank) {        // Previous ranks were bad
                nbr = 0;
                rank_wrong = r_w;
                rank = rr;
            } else if (rr == rank) { // another rank
                rank_wrong = (r_w)*(rank_wrong);
            } else               // bad rank (happens with proba wrong)
                --nbr;
        }
        int_wrong = 1;
    }


    wrong = int_wrong;
}

   




#endif 
