// =========================================================
// (C) The Linbox Group 1999
// Linbox wrapper for sparse vectors
// file : lin_dom_spv_bb.h
// Time-stamp: <08 Dec 11 12:18:05 Jean-Guillaume.Dumas@imag.fr> 
// =========================================================
#ifndef __SPARSE_ERASE_B_B_DOMAIN_H__
#define __SPARSE_ERASE_B_B_DOMAIN_H__

#include "LinBox/lin_sparse_vector.h"
#include "LinBox/lin_spv_bb.h"

template <class Domain, class I = unsigned long>
class SparseBlackBoxEraseDom : public SparseBlackBoxDom<Domain, I>  {
public:
    typedef          SparseBlackBoxDom< Domain, I >                  Father_t;
    using Father_t::_domain;
    using Father_t::_container;
    using Father_t::_row_dim;
    using Father_t::_col_dim;
    using Father_t::_nz_elem;
    typedef typename Father_t::SV_t SV_t;
    typedef typename Father_t::Row_t Row_t;
    typedef typename Father_t::Type_t      Type_t;
    typedef typename Father_t::Rep         Rep;
public:
    typedef          SparseBlackBoxEraseDom< Domain, I >           Self_t;
public:
        ///-- Default cstors:
    SparseBlackBoxEraseDom() : Father_t() {};
    SparseBlackBoxEraseDom(const Domain& D) : Father_t(D) {}
    SparseBlackBoxEraseDom(const Domain& D, char * mat_file) : Father_t(D, mat_file) {}
    
    
        ///-- Cstor of recopy: compiler's generated
    SparseBlackBoxEraseDom(const Self_t& M) : Father_t( static_cast<Father_t>(M) ) {}

   unsigned long init_erase(Rep& a, unsigned long& ni, unsigned long& nj, char * mat_file) const {
        unsigned long ne;  
        return read_erase(a,ni,nj,ne,mat_file); 
    }    

  // ***********************************************************
  // Access to the sparse matrix representation of the black-box 
  // Reads to a file in the sparse format if the entries can be 
  // converted to %ld, otherwise ?
    unsigned long read_erase (Rep& ca, unsigned long& ni, unsigned long& nj, unsigned long& ne, char * mat_file) const { 
        char *UT, *File_Name;
        int is_gzipped = 0;
        size_t s = strlen(mat_file);
        if ((mat_file[--s] == 'z') && (mat_file[--s] == 'g') && (mat_file[--s] == '.')) {
            is_gzipped = 1;
            File_Name = tempnam("/tmp","bbx_"); // "/tmp/bbx_XXXXXX"; mkstemp(File_Name);
            UT = new char[s+34+strlen(File_Name)];
            sprintf(UT,"gunzip -c %s > %s", mat_file, File_Name);
            system(UT);
            sprintf(UT,"\\rm %s", File_Name);
        } else
            File_Name = mat_file;

        FILE* FileDes = fopen(File_Name, "r");
        unsigned long erased = 0, block = 0;
        if (FileDes != NULL) {
 	    char * tmp = new char[80];   
            fscanf(FileDes,"%ld %ld %s\n",&ni, &nj, tmp) ;
//             ca = Rep( ni ); ne=0;
            ca.resize( ni ); ne=0;

            long i,j, val;
            fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
//             typename Domain_t::element cour;
            Type_t cour;
            std::vector<long> removecolumns;
           
            for(long ii=0; ii<(long)ni; ++ii) {
                    // No non-zero element yet
//                 ca[ii] = SV_t(0,nj);
                ca[ii].resize(0); ca[ii].reactualsize(nj);
                while (i == (ii+(long)erased+1L)) {
                    _domain.init( cour, val );
                    if (! _domain.isZero( cour )) {
                        if (removecolumns.size()) {
                            std::vector<long>::const_iterator vvit = removecolumns.begin();
                            for(; vvit != removecolumns.end(); ++vvit) {
                                if (*vvit == j) break;
                                else if (*vvit > j) {
                                    vvit = removecolumns.end();
                                    break;
                                }
                            }
                            if (vvit == removecolumns.end()) {
                                ca[ii].push_back( typename SV_t::value_type(j-1, cour ) );
                                ++ne;
                            }
                        } else {
                            ca[ii].push_back( typename SV_t::value_type(j-1, cour ) );
                            ++ne;
                        }
                    }
                    fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
                }
                    // Erasing phase
                    // 0 lines are erased
                if (ca[ii].size() == 0) {
                    ++erased;--ii;--ni;
                }
                    // Singleton lines are erased, rank is incremented
                if (ca[ii].size() == 1) {
                    ++block;
                    long rj = ca[ii][0].getindex();
                    if (removecolumns.size()) {
                        std::vector<long>::iterator rcit = removecolumns.begin();
                        for(; rcit != removecolumns.end(); ++rcit) {
//                             if (*rcit == (rj+1)) {
//                                 --block;
//                                 break;
//                             }
                            if (*rcit > rj) {
                                removecolumns.insert(rcit, rj+1);
                                break;
                            }
                        }
                        if (rcit == removecolumns.end()) removecolumns.push_back(rj+1);
                    } else {
                        removecolumns.push_back(rj+1);
                    }
                    for(long newi=0; newi<ii; ++newi) {
                        for(typename Row_t::iterator rowit = ca[newi].begin();rowit != ca[newi].end(); ++rowit) {
                            if ( static_cast<long>((*rowit).getindex()) == rj) {
                                ca[newi].erase(rowit);
                                break;
                            } else { 
                                if (static_cast<long>((*rowit).getindex()) > rj)
                                    break;
                            }
                        }
                    }
                    ++erased;--ii;--ni;
                }
            }
            ca.resize(ni);

            fclose(FileDes);
            if (is_gzipped) {
                system(UT);
            }        
        }

        return block;
    }
     

    unsigned long read_erase (char * mat_file) { return read_erase(_container,_row_dim,_col_dim,_nz_elem, mat_file); }
    
    unsigned long read_erase (unsigned long& ni, unsigned long& nj, unsigned long& ne, char * mat_file) { unsigned long addrank = read_erase(_container,_row_dim,_col_dim,_nz_elem, mat_file); ni = _row_dim; nj = _col_dim; ne = _nz_elem; return addrank; }
    

};



#endif // __SPARSE_ERASE_B_B_DOMAIN_H__
